// OpenSSL.cpp : Definiert die Initialisierungsroutinen fr die DLL.
//

#include "stdafx.h"
#include "OpenSSL.h"
// DllGetVersion
#include "shlwapi.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#endif

//
//TODO: Wenn diese DLL dynamisch mit MFC-DLLs verknpft ist,
//		muss fr alle aus dieser DLL exportierten Funktionen, die in
//		MFC aufgerufen werden, das AFX_MANAGE_STATE-Makro
//		am Anfang der Funktion hinzugefgt werden.
//
//		Beispiel:
//
//		extern "C" BOOL PASCAL EXPORT ExportedFunction()
//		{
//			AFX_MANAGE_STATE(AfxGetStaticModuleState());
//			// Hier normaler Funktionsrumpf
//		}
//
//		Es ist sehr wichtig, dass dieses Makro in jeder Funktion
//		vor allen MFC-Aufrufen angezeigt wird. Dies bedeutet,
//		dass es als erste Anweisung innerhalb der 
//		Funktion angezeigt werden muss, sogar vor jeglichen Deklarationen von Objektvariablen,
//		da ihre Konstruktoren Aufrufe in die MFC-DLL generieren
//		knnten.
//
//		Siehe Technische Hinweise fr MFC 33 und 58 fr weitere
//		Details.
//

// COpenSSLApp

BEGIN_MESSAGE_MAP(COpenSSLApp, CWinApp)
END_MESSAGE_MAP()


// COpenSSLApp-Erstellung

// Forward deklaration
extern "C"
int WINAPI SSLDllGetVersionString(char*, int);

COpenSSLApp::COpenSSLApp() : CLog(LOG_CHANNEL_API)
{
	// Member initialisieren
	m_ctx = NULL;
	m_wndPlugIn = NULL;
	// Version der DLL ermitteln.
	char version[32];
	memset(version, 0, sizeof(version));
	SSLDllGetVersionString(version, sizeof(version));
	// Laden des Logs
	logIniRead();
	logBeginEnd("INFO: Logging started (DLL-Version: %s).", version);
}

COpenSSLApp::~COpenSSLApp()
{
	// Beenden des logs.
	logBeginEnd("INFO: Logging stopped.");
}

// Das einzige COpenSSLApp-Objekt

COpenSSLApp theApp;

// COpenSSLApp-Initialisierung

BOOL COpenSSLApp::InitInstance()
{
	CWinApp::InitInstance();

	// WinSockets starten
	WSADATA wsadata;
	WSAStartup(MAKEWORD(2, 2), &wsadata);

	//////////////////////////////////////////////////////////
	// PlugIn Vorbereitung

	// Wnd erzeugen
	m_wndPlugIn = new CWndPlugIn();

	if (!m_wndPlugIn->CreateEx(0, "STATIC", "WndPlugIn", WS_OVERLAPPEDWINDOW, 0,0,10,10, NULL, NULL)) {

		logBeginEnd("InitInstance: Could not create WndPlugIn.");
		delete m_wndPlugIn;
		m_wndPlugIn = NULL;
		return FALSE;
	}
	// Fenster nicht anzeigen
	m_wndPlugIn->ShowWindow(FALSE);

	// Fenster an Anwendungsobjekt binden
	m_pMainWnd = m_wndPlugIn;

	//  Nachrichtensystem der Anwendung starten.
	return TRUE;
}

int COpenSSLApp::ExitInstance()
{
	// Wnd freigeben.
	if (m_wndPlugIn != NULL) {

		m_wndPlugIn->DestroyWindow();
		delete m_wndPlugIn;
		m_wndPlugIn = NULL;
		m_pMainWnd = NULL;
	}	

	// WinSockets beenden
	WSACleanup();

	return CWinApp::ExitInstance();
}

/////////////////////////////////////////////////
//################# API #########################
// ########################################################
/**
* @fn    DllGetVersion(DLLVERSIONINFO *pdvi)
* @brief Ermitteln der DLL - Version. Kompatible zu Windows - Standard.
*
* @param pdvi       { Zeiger auf Struktur zur Informationsuebergabe. }
*/
extern "C" 
HRESULT WINAPI SSLDllGetVersion(DLLVERSIONINFO *pdvi)
{
	char szModuleName[MAX_PATH];
	DWORD dwHandle;
	DWORD dwLen;
	LPVOID lpBuffer = NULL;
	UINT uiLen;

	unsigned char *pucData = NULL;

	// zulaessige Bereiche pruefen
	if (pdvi == NULL) {

		return E_ABORT;
	}

	if (pdvi->cbSize != sizeof(DLLVERSIONINFO)) {

		return E_ABORT;
	}
	// Modulname der DLL
	if (!GetModuleFileName(theApp.m_hInstance, szModuleName, MAX_PATH)) {

		return E_ABORT;
	}
	// Laenge der Version - Resource
	dwLen = GetFileVersionInfoSize(szModuleName, &dwHandle);
	if (!dwLen) {

		return E_ABORT;
	}

	pucData = new unsigned char[dwLen + 100]; // unicode
	// DLL - Versioninfo ermitteln
	if (!GetFileVersionInfo(szModuleName, dwHandle, dwLen, (LPVOID)pucData)) {

		delete pucData;
		return E_ABORT;
	}
	// Pointer auf allgemeine Infostruktur ermitteln
	if (!VerQueryValue((LPVOID)pucData, "\\\0", &lpBuffer, &uiLen)) {

		delete pucData;
		return E_ABORT;
	}
	// Daten zurueckgeben
	VS_FIXEDFILEINFO *pvs_ffi = (VS_FIXEDFILEINFO*)lpBuffer;
	pdvi->dwMajorVersion = pvs_ffi->dwFileVersionMS >> 16;
	pdvi->dwMinorVersion = pvs_ffi->dwFileVersionMS & 0x0000FFFF;
	pdvi->dwBuildNumber  = pvs_ffi->dwFileVersionLS >> 16;
	pdvi->dwPlatformID   = DLLVER_PLATFORM_WINDOWS;

	delete pucData;

	return NOERROR;
}

extern "C" 
int WINAPI SSLDllGetVersionString(char *version, int size)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	DLLVERSIONINFO vi;
	memset(&vi, 0, sizeof(DLLVERSIONINFO));
	vi.cbSize = sizeof(DLLVERSIONINFO);
	if (SSLDllGetVersion(&vi) == NOERROR) {

		return _snprintf(version, size, "%02i.%02i.%02i", vi.dwMajorVersion, vi.dwMinorVersion, vi.dwBuildNumber);
	}

	return 0;
}

extern "C" 
int WINAPI SSLStartup(HWND hPlugInWnd)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	unsigned long err;
	char err_buf[128];
	int i, r;

	//////////////////////////////////////////////////////////
	// PlugIn - Vorbereitung

	// Eigenes Fensterhandle Prfen
	if (theApp.m_wndPlugIn == 0) {

		theApp.logBeginEnd("SSLStartup [Wrong PlugIn-WndHandle or couldn't extract.]");
		return 0;
	}

	// bergebenen Fensterhandle Prfen
	if (hPlugInWnd == 0 || !::IsWindow(hPlugInWnd)) {

		theApp.logBeginEnd("SSLStartup [Wrong Nettalk-WndHandle or couldn't extract.]");
		return 0;
	}

	// Fensterhandle von Nettalk an PlugIn-Fenster weiterreichen. PlugIn installieren.
	theApp.m_wndPlugIn->InstPlugIn(hPlugInWnd);

	//////////////////////////////////////////////////////////
	// OpenSSL - Bibliothek Vorbereitung

	// Initialisieren der SSL Bibliotek
	SSL_library_init();
	SSL_load_error_strings();

	theApp.m_ctx = SSL_CTX_new(SSLv23_client_method());
	if (theApp.m_ctx == 0) {

		if ((err = ERR_get_error()) > 0) {

			ERR_error_string (err, err_buf);
			theApp.logBeginEnd("SSLStartup: [SSL_CTX_new](%d): %s", err, err_buf);
		}
		// Megafatal, aber abgefangen!
		theApp.logBeginEnd("SSLStartup [SSL_CTX_new] fail.");
		return 0;
	}

	// Initialisieren des PRNG
	for (i = 0; i < 128; i++) {

		r = rand();
		RAND_seed((unsigned char *)&r, sizeof (r));
	}

	return 1;
}

// Forward fr Disconnect bei ExitInstance
extern "C" 
int WINAPI SSLDisconnect(SOCKET);

extern "C" 
int WINAPI SSLCleanup()
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	// Noch offene Verbindungen?
	if (theApp.m_mapSSLSocketData.GetSize()) {

		// Alle offenen Verbindungen runterfahren.
		CSSLSocketData *sd;
		SOCKET sock;
		POSITION pos = theApp.m_mapSSLSocketData.GetStartPosition();
		while (pos) {

			theApp.m_mapSSLSocketData.GetNextAssoc(pos, sock, sd);
			SSLDisconnect(sock);
		}
		// Alles bereinigt.
		theApp.m_mapSSLSocketData.RemoveAll();
	}

	// Context erstellt?
	if (theApp.m_ctx != NULL) {

		// App-Context Cleanup
		SSL_CTX_free(theApp.m_ctx);
	}

	// Alle weiteren erdenklichen Resourcen freigeben.
	ERR_remove_state(0);
	ERR_free_strings();
	//ENGINE_cleanup();
	//CONF_modules_unload();
	EVP_cleanup();
	CRYPTO_cleanup_all_ex_data();
	RAND_cleanup();

	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
	// brige Memory leaks bitte verzeihen, is ja scheinbar nach docu net schlimm :o)-
	//++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

	// PlugIn
	if (theApp.m_wndPlugIn != NULL) {

		// PlugIn abmelden
		theApp.m_wndPlugIn->UnloadPlugIn();
	}

	return 1;
}

static void broke_oneline(char *oneline, char *parray[])
{
	char *pt, *ppt;
	int i;

	i = 0;
	ppt = pt = oneline + 1;
	while ((pt = strchr (pt, '/'))) {

		*pt = 0;
		parray[i++] = ppt;
		ppt = ++pt;
	}
	parray[i++] = ppt;
	parray[i] = NULL;
}

static void ASN1_TIME_snprintf(char *buf, int buf_len, ASN1_TIME *tm)
{
	char *expires = NULL;
	BIO *inMem = BIO_new(BIO_s_mem());

	ASN1_TIME_print(inMem, tm);
	BIO_get_mem_data(inMem, &expires);
	buf[0] = 0;
	if (expires != NULL) {

		memset(buf, 0, buf_len);
		strncpy(buf, expires, 24);
	}
	BIO_free(inMem);
}

void PrintCertificateInfo(CSSLSocketData &sd)
{
	X509 *peer_cert;
	EVP_PKEY *peer_pkey;
	char notBefore[64];
	char notAfter[64];
	int i, alg, sign_alg;
	struct cert_info cert_info;
	char buf[128];

	if (!(peer_cert = SSL_get_peer_certificate (sd.m_ssl))) {

		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", "- No Certificate");
		return;
	}

	X509_NAME_oneline(X509_get_subject_name(peer_cert), cert_info.subject, sizeof (cert_info.subject));
	X509_NAME_oneline(X509_get_issuer_name(peer_cert), cert_info.issuer, sizeof (cert_info.issuer));
	broke_oneline(cert_info.subject, cert_info.subject_word);
	broke_oneline(cert_info.issuer, cert_info.issuer_word);

	alg = OBJ_obj2nid(peer_cert->cert_info->key->algor->algorithm);
	sign_alg = OBJ_obj2nid(peer_cert->sig_alg->algorithm);
	ASN1_TIME_snprintf(notBefore, sizeof (notBefore), X509_get_notBefore(peer_cert));
	ASN1_TIME_snprintf(notAfter, sizeof (notAfter), X509_get_notAfter(peer_cert));

	peer_pkey = X509_get_pubkey(peer_cert);

	strncpy(cert_info.algorithm, (alg == NID_undef) ? "Unknown" : OBJ_nid2ln (alg), sizeof (cert_info.algorithm));
	cert_info.algorithm_bits = EVP_PKEY_bits (peer_pkey);
	strncpy(cert_info.sign_algorithm, (sign_alg == NID_undef) ? "Unknown" : OBJ_nid2ln (sign_alg), sizeof (cert_info.sign_algorithm));
	/* EVP_PKEY_bits(ca_pkey)); */
	cert_info.sign_algorithm_bits = 0;
	strncpy(cert_info.notbefore, notBefore, sizeof (cert_info.notbefore));
	strncpy(cert_info.notafter, notAfter, sizeof (cert_info.notafter));

	EVP_PKEY_free(peer_pkey);

	/* SSL_SESSION_print_fp(stdout, SSL_get_session(ssl)); */
	/*
	if (ssl->session->sess_cert->peer_rsa_tmp) {
	tmp_pkey = EVP_PKEY_new();
	EVP_PKEY_assign_RSA(tmp_pkey, ssl->session->sess_cert->peer_rsa_tmp);
	cert_info.rsa_tmp_bits = EVP_PKEY_bits(tmp_pkey);
	EVP_PKEY_free(tmp_pkey);
	} else
	fprintf(stderr, "REMOTE SIDE DOESN'T PROVIDES ->peer_rsa_tmp\n");
	*/
	cert_info.rsa_tmp_bits = 0;

	X509_free(peer_cert);

	snprintf (buf, sizeof (buf), "- Certification info:");
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	snprintf(buf, sizeof (buf), "  Subject:");
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	for (i = 0; cert_info.subject_word[i]; i++)	{

		snprintf (buf, sizeof (buf), "    %s", cert_info.subject_word[i]);
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);
	}

	snprintf (buf, sizeof (buf), "  Issuer:");
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	for (i = 0; cert_info.issuer_word[i]; i++) {

		snprintf (buf, sizeof (buf), "    %s", cert_info.issuer_word[i]);
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);
	}

	snprintf (buf, sizeof (buf), "  Public key algorithm: %s (%d bits)", cert_info.algorithm, cert_info.algorithm_bits);
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	/*if (cert_info.rsa_tmp_bits)
	{
		snprintf (buf, sizeof (buf), "  Public key algorithm uses ephemeral key with %d bits",
		cert_info.rsa_tmp_bits);
	}*/

	snprintf (buf, sizeof (buf), "  Sign algorithm %s", cert_info.sign_algorithm/*, cert_info.sign_algorithm_bits*/);
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	snprintf (buf, sizeof (buf), "  Valid since %s to %s", cert_info.notbefore, cert_info.notafter);
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	struct chiper_info chiper_info;
	SSL_CIPHER *c = SSL_get_current_cipher(sd.m_ssl);
	strncpy (chiper_info.version, SSL_CIPHER_get_version(c), sizeof(chiper_info.version));
	strncpy (chiper_info.chiper, SSL_CIPHER_get_name(c), sizeof(chiper_info.chiper));
	SSL_CIPHER_get_bits(c, &chiper_info.chiper_bits);

	snprintf (buf, sizeof (buf), "- Cipher info:");
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	snprintf (buf, sizeof (buf), "  Version: %s, cipher %s (%u bits)", chiper_info.version, chiper_info.chiper, chiper_info.chiper_bits);
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);

	int verify_error = SSL_get_verify_result(sd.m_ssl);
	switch (verify_error)
	{
	case X509_V_OK:
		/* snprintf (buf, sizeof (buf), "* Verify OK (?)"); */
		/* EMIT_SIGNAL (XP_TE_SERVTEXT, serv->server_session, buf, NULL, NULL, NULL, 0); */
		break;
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
	case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
	case X509_V_ERR_CERT_HAS_EXPIRED:
		if (sd.m_acceptInvalideCertificate)
		{
			snprintf (buf, sizeof (buf), "- Verify E: %s.? (%d) -- Ignored", X509_verify_cert_error_string(verify_error), verify_error);
			theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);
			break;
		}
	default:
		snprintf (buf, sizeof (buf), "- %s.? (%d)", X509_verify_cert_error_string(verify_error), verify_error);
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", buf);
	}
}

#undef FUNC_NAME
#define FUNC_NAME ""

extern "C" 
int WINAPI SSLConnect(SOCKET sock, int serverFrameID, int acceptInvalidCertificate, char *hostName)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	unsigned long err;
	int res;
	char err_buf[128];
	char buf[128];
	int retries = 0;

	// Werte an SSLSocket bergeben.
	CSSLSocketData &sd = theApp.SSLSocketDataRequest(sock);
	sd.m_serverFrameId = serverFrameID;
	sd.m_acceptInvalideCertificate = acceptInvalidCertificate;
	strncpy(sd.m_hostName, hostName, sizeof(sd.m_hostName) - 1);

	// Erzeugen des Hauptkontext, sollte immer ok sein!
	SSL *ssl = SSL_new(theApp.m_ctx);
	if (ssl == 0) {
		if ((err = ERR_get_error()) > 0) {

			ERR_error_string(err, err_buf);
			theApp.logBeginEnd("SSL_new(%d) %s", err, err_buf);
		}
		theApp.logBeginEnd("[SSL_new] fail.");
		// Fatal
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", "- [OpenSSL] Connect");
		theApp.m_mapSSLSocketData.RemoveKey(sock);
		return 0;
	}

	// Verbindungsstrukt halten
	sd.m_ssl = ssl;

	// Verbundenen Socket mit SSL verknpfen.
	if (SSL_set_fd(sd.m_ssl, sock) == 0) {
		
		// Fehler bei Verknpfung von Socket mit SSL
		SSL_free(sd.m_ssl);
		theApp.logBeginEnd("[SSL_set_fd] fail.");
		// Fatal
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", "- [OpenSSL] Connect");
		theApp.m_mapSSLSocketData.RemoveKey(sock);
		return 0;
	}

retry_connect:
	if ((res = SSL_connect(sd.m_ssl)) <= 0) {

		ERR_error_string(res, err_buf);
		theApp.logBeginEnd("SSL_connect error (%d): %s", res, err_buf);

		// Zu viele Fehler am Stck lsst die Routine abbrechen.
		if (++retries < 11) { // Worst case: 10 * 500ms = 5 Sekunden lang Fehlzustand (nicht akzeptabel).

			if ((res = SSL_get_error(sd.m_ssl, res)) > 0) {

				ERR_error_string (res, err_buf);
				theApp.logBeginEnd("SSL_connect error (%d): %s", res, err_buf);

				switch (res) {
				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_CONNECT:
				case SSL_ERROR_WANT_ACCEPT:
				case SSL_ERROR_WANT_X509_LOOKUP:
				case SSL_ERROR_WANT_READ:{
					Sleep(500);
					goto retry_connect;}
				case SSL_ERROR_SYSCALL: {
					if (ERR_get_error() == 0) {
						Sleep(500);
						goto retry_connect;
					}}
				default:
					// Fehler bei Verbindungsaufbau.
					theApp.logBeginEnd("[SSL_connect] fail.");
				}
			}
		}

		// Fataler Fehler, kann nicht fortgesetzt werden.
		SSL_free(sd.m_ssl);
		// Fatal
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", "- [OpenSSL] Connect");
		theApp.m_mapSSLSocketData.RemoveKey(sock);
		return 0;
	}

	int verify_error = SSL_get_verify_result(sd.m_ssl);
	switch (verify_error)
	{
	case X509_V_OK:
		/* snprintf (buf, sizeof (buf), "* Verify OK (?)"); */
		/* EMIT_SIGNAL (XP_TE_SERVTEXT, serv->server_session, buf, NULL, NULL, NULL, 0); */
		break;
	case X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY:
	case X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE:
	case X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT:
	case X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN:
	case X509_V_ERR_CERT_HAS_EXPIRED:
		if (sd.m_acceptInvalideCertificate)
		{
			snprintf (buf, sizeof (buf), "- Verify E: %s.? (%d) -- Ignored", X509_verify_cert_error_string(verify_error), verify_error);
			theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", buf);
			break;
		}
	default:
		snprintf (buf, sizeof (buf), "- Verify E: %s.? (%d)", X509_verify_cert_error_string(verify_error), verify_error);
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", buf);
		SSL_free(sd.m_ssl);
		theApp.m_mapSSLSocketData.RemoveKey(sock);
		return 0;	
	}

	// Zusatzprfung auf Hostnamen durchfhren
	if (sd.m_acceptInvalideCertificate == 0) {

		char peer_CN[256];
		memset(peer_CN, 0, sizeof(peer_CN));

		/*Check the cert chain. The chain length
		is automatically checked by OpenSSL when
		we set the verify depth in the ctx */

		/*Check the common name*/
		X509 *peer_cert = SSL_get_peer_certificate(sd.m_ssl);
		X509_NAME_get_text_by_NID(X509_get_subject_name(peer_cert), NID_commonName, peer_CN, 256);

		if(strcmp(peer_CN, sd.m_hostName)) {

			theApp.logBeginEnd("Common name doesn't match host name.");
			theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$04>", "- [OpenSSL] Common name doesn't match host name, Certificate");
			SSL_free(sd.m_ssl);
			theApp.m_mapSSLSocketData.RemoveKey(sock);
			return 0;
		}
	}

	// Zusatzinfos ausgeben
	//PrintCertificateInfo(sd);

	theApp.logBeginEnd("SSLConnect durchgefhrt.");
	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo <$03>", "- [OpenSSL] Connect");
	return 1;
} 

#undef FUNC_NAME
#define FUNC_NAME ""

extern "C" 
int WINAPI SSLDisconnect(SOCKET sock)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int res, retries = 0;
	char err_buf[128];
	memset(err_buf, 0, sizeof(err_buf));

	CSSLSocketData &sd = theApp.SSLSocketDataRequest(sock);
	if (sd.m_ssl != NULL) {

	// SSL Verbindung beenden
retry_shutdown:
		if ((res = SSL_shutdown(sd.m_ssl)) < 0) {

			ERR_error_string(res, err_buf);
			theApp.logBeginEnd("SSL_shutdown error (%d): %s", res, err_buf);

			if ((res = SSL_get_error(sd.m_ssl, res)) > 0) {

				ERR_error_string(res, err_buf);
				theApp.logBeginEnd("SSL_shutdown error (%d): %s", res, err_buf);

				switch (res) {
				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_READ:{
					Sleep(250);
					if (++retries > 5) {

						theApp.logBeginEnd("[SSL_shutdown maximal tries -> %i reached].", retries);
						return 0;
					}
					goto retry_shutdown;}

				default:
					theApp.logBeginEnd("[SSL_shutdown] fail.");
					theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo", "- [OpenSSL] Disconnect fail...");
				}
			}
		}

		// Fatal oder nicht, Verbindung freigeben.
		SSL_free(sd.m_ssl);
		theApp.m_mapSSLSocketData.RemoveKey(sock);
		return 0;
	} else {

		// Aufruf obwohl keine Verbindung
		theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo", "- [OpenSSL] Not connected...");
		return 0;

	}

	theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo", "- [OpenSSL] Disconnect ok...");
	return 1;
} 

#undef FUNC_NAME
#define FUNC_NAME ""

extern "C" 
int WINAPI SSLSend(SOCKET sock, unsigned char *buf, int num)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int res, retries = 0;
	char err_buf[128];
	memset(err_buf, 0, sizeof(err_buf));

	CSSLSocketData &sd = theApp.SSLSocketDataRequest(sock);
	if (sd.m_ssl == NULL) {

		theApp.logBeginEnd("SSL_write and not connected.");
		return 0;
	}
retry_write:
	if ((res = SSL_write(sd.m_ssl, buf, num)) <= 0) {

		theApp.logBeginEnd("SSL_write fail res(%i)", res);

		if ((res = SSL_get_error(sd.m_ssl, res)) > 0) {

			theApp.logBeginEnd("SSL_write fail get_error(%i)", res);

			ERR_error_string(res, err_buf);
			theApp.logBeginEnd("SSL_write error (%d): %s", res, err_buf);

			switch (res) {
				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_CONNECT:
				case SSL_ERROR_WANT_ACCEPT:
				case SSL_ERROR_WANT_X509_LOOKUP:
				case SSL_ERROR_WANT_READ:{
					Sleep(250);
					if (++retries > 5) {

						theApp.logBeginEnd("[SSL_write maximal tries -> %i reached].", retries);
						return 0;
					}
					goto retry_write;}

				default:
					// Fataler Fehler abbruch ntig.
					theApp.logBeginEnd("[SSL_write switch->default] fail.");
					theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo", "- [OpenSSL] Send fail...");
					return SOCKET_ERROR;
			}
		}

		return SOCKET_ERROR;
	}

	theApp.logBeginEnd("SSLSend return(%i)", res);
	return res;
} 

#undef FUNC_NAME
#define FUNC_NAME ""

extern "C" 
int WINAPI SSLRecv(SOCKET sock, unsigned char *buf, int num)
{
	AFX_MANAGE_STATE(AfxGetStaticModuleState());

	int res, retries = 0;
	char err_buf[128];
	memset(err_buf, 0, sizeof(err_buf));

	CSSLSocketData &sd = theApp.SSLSocketDataRequest(sock);
	if (sd.m_ssl == NULL) {

		theApp.logBeginEnd("SSL_read and not connected...");
		return 0;
	}
retry_read:
	if ((res = SSL_read(sd.m_ssl, buf, num)) <= 0) {

		theApp.logBeginEnd("SSL_read fail res(%i)", res);

		if ((res = SSL_get_error(sd.m_ssl, res)) > 0) {

			theApp.logBeginEnd("SSL_read fail get_error(%i)", res);

			ERR_error_string(res, err_buf);
			theApp.logBeginEnd("SSL_read error (%d): %s", res, err_buf);

			switch (res) {
				case SSL_ERROR_WANT_WRITE:
				case SSL_ERROR_WANT_CONNECT:
				case SSL_ERROR_WANT_ACCEPT:
				case SSL_ERROR_WANT_X509_LOOKUP: {
					Sleep(250);
					if (++retries > 3) {

						theApp.logBeginEnd("[SSL_read maximal tries -> %i reached].", retries);
						return 0;
					}
					goto retry_read;}
				case SSL_ERROR_WANT_READ: {

					fd_set readfds;
					struct timeval seltime;
					FD_ZERO((fd_set FAR *) &readfds);
					FD_SET(sock, (fd_set FAR *) &readfds);
					seltime.tv_sec  = 1;
					seltime.tv_usec = 0;

					if ((res = select(FD_SETSIZE,
						(fd_set FAR *)&readfds, (fd_set FAR *) 0, (fd_set FAR *) 0,
						(CONST struct timeval FAR *)&seltime)) == SOCKET_ERROR) {

							theApp.logBeginEnd("[SSL_read socketerror(%i) ]", WSAGetLastError());
							return 0;
					} else if (res != 0) {
						// Es liegen Daten an.
						goto retry_read;
					}
					break;}

				default:
					theApp.logBeginEnd("[SSL_read switch->default] fail.");
					theApp.m_wndPlugIn->SendCommand(sd.m_serverFrameId, "/echo", "- [OpenSSL] Recv fail...");
					return SOCKET_ERROR;
			}
		}

		theApp.logBeginEnd("[SSL_read SOCKET_ERROR] fail.");
		return SOCKET_ERROR;
	}

	if (res >= 0) {

		buf[res] = 0;
	}

	theApp.logBeginEnd("SSLRecv return(%i)", res);
	return res;
} 
